# coding: utf-8
"""
GUI base classes
"""
from copy import deepcopy

import pygame
import pygame.locals as loc

from style import Style

class Widget(object):
    """
    Base class of all GUI elements.

    """
    #: List of valid anchors. Do not change or edit.
    _anchors = ("topleft", "bottomleft", "topright", "bottomright",
                "midtop", "midleft", "midbottom", "midright", "center")

    #: If True, the widget will pass all events. This means, events may be
    #: used more than once. Defaults to False.
    pass_events = False
    #: If False, the widget will be unable to take the :ref:`focus <focus>`\ . 
    #: Otherwise, :attr:`focused` will be set to True if the widget
    #: was clicked. Defaults to True.
    focusable = True
    #: Class name that will be used to get per-class styles (case insensitive). 
    #: If it is ``None`` (which is the default), the class name will be used.
    #: This may be useful for inherited classes that should belong to the same
    #: style class group like their parent class. See :ref:`Styling`
    #: .. note:: 
    #:    
    #:     Since this is a normal class attribute, this will be inherited
    #:     to all child classes. So only use when required.
    #:
    style_class_name = None
    #: This should be a complete list of names of style attributes that
    #: you will use. If this is too generic, overwrite
    #: :meth:`Widget.get_styles`
    used_styles = ["image", "bgcolor"]
    highlight = False
    def __init__(self, parent, size, pos, anchor="topleft",
                 index="end", style={}, **kwargs):
        """
        :type parent: :class:`Layer`
        :param parent: Parent widget
        
        :type size: :ref:`Size`
        :param size:
              Size of the widget.

        :type pos: :ref:`Position`
        :param pos: 
            Position of the widget. (Note the :attr:`anchor` attribute, too.)
        
        :type anchor: String
        :param anchor:
            The point that should be used for positioning. Must be one of the
            ``pygame.Rect`` attributes (topleft, midtop, center, ...)
        
        :type index: int or "end"
        :param index:
            Index that will be used when adding the widget to the parent
            widget (via :meth:`Layer.add`\ ). The rearmost position is at
            index 0, the topmost position at ``"end"`` (default).
        
        
        :type visible: boolean
        :param visible: 
            Whether the widget should be visible or not. Defaults to ``True``.
            
        :type validate_pos: boolean
        :param validate_pos:
            If True, the widget cannot tower above the parent. This will
            affect moving and resizing of the widget.
              
        :type min_size: :ref:`Size`
        :param min_size: 
            Minimum size for the widget. Defaults to ``("1px, "1px")``\ .
        
        :type max_size: :ref:`Size`
        :param max_size: 
            Maximum size for the widget. Defaults to ``("100%", "100%")``\ .
            
        :type style: :class:`Style` or dict
        :param style:
            Widget specific style attributes. Fallback defaults to
            the main GUI style.
            
        :type style_class_name: String
        :param style_class_name:
            Prefix used by Style when searching for a style attribute.
            See documentation on :ref:`styling`\ .
            
        :type highlight: boolean
        :param highlight:
            If True, style names will be suffixed with ``_active`` on mouse
            over or dragging. Defaults to False.
        
        """
        # initialized flag
        self.init = False
        # Variables to store values for the :attr:`pos` and :attr:`size`
        # properties. Do not access directly.
        #: internal variable to store the position
        self._pos = [(1, "px", 0), (1, "px", 0)]
        #: internal variable to store the size
        self._size = [(1, "px", 0), (1, "px", 0)]

        #: event names -> lists of event listeners. Do not access directly.
        self.listeners = {}
        
#        #: list of dirty rects. This is useful if the position or size
#        #: changes and the old rect needs to be updated. Thus, for all methods
#        #: that change the blitting behavior outside of the actual :meth:`blit`
        self.drects = []
        #: dirty flag, set to force reblit (normally not necessary)
        self.dirty = True
        
        #: internally used to cache styles and detect style changes
        self._cached_styles = {}

        #: reference to the parent widget
        self.parent = parent

        # a hook to allow the GUI widget to have no parent
        if self.parent is not None:
            self.parent.add(self)

        #: anchor to use when calculating the widget position. Must be one
        #: of the pygame.Rect attributes.
        self.anchor = anchor

        # ensure that :attr:`anchor` is is a valid pygame.Rect attribute name
        if anchor not in self._anchors:
            raise ValueError, 'invalid anchor "%s": must be one of %s' % (
                             anchor, ", ".join(self._anchors))
        
        if "style_class_name" in kwargs:
            self.style_class_name = kwargs["style_class_name"]
        if isinstance(style, Style): # and not style.fallback:
            #style.fallback = default_fallback
            self.style = style
        elif isinstance(style, dict):
            default_fallback = self.get_gui().style
            self.style = Style(fallback=default_fallback, **style)
        else:
            raise ValueError, 'invalid style "%s": must be Style or dict.' % (
                                style)
        
        # some more attributes
        self.visible = kwargs.get("visible", True)
        self.validate_pos = kwargs.get("validate_pos", False)
        if self.validate_pos:
            default_max_size = ["100%", "100%"] 
        else: 
            default_max_size = [10000, 10000]
        self.max_size = kwargs.get("max_size", default_max_size)
        self.min_size = kwargs.get("min_size", ["1px", "1px"])
        
        # internal variables to cache position and size
        self.update_pos_cache = True
        self._cached_pos = [0, 0]
        self.update_size_cache = True
        self._cached_size = [1, 1]

        self.local_pos = [0, 0]
        
        self.pos = pos
        self.size = size
        
        #: surface to cache last widget blit
        self.image = pygame.Surface(self.size, loc.SRCALPHA)
        
        # if not in kwargs, use the class default value
        self.pass_events = kwargs.get("pass_events", self.pass_events)
        
        self.highlight = kwargs.get("highlight", self.highlight)
        
        self.focusable = kwargs.get("focusable", self.focusable)
        
        # some flags and variables for the internal event system
        
        # widget is focused
        self.focused = False
        # mouse is over the widget
        self.mouse_over = False
        # last absolute mouse position on the widget
        self.last_mouse_pos = None
        # whether the mouse is dragging with various buttons
        # will be set to 0 on "mouse_left"
        self.dragging = [0, 0, 0]
        # point relative to the own topleft where the mouse was pressed
        self.dragging_anchor = None
        # last mouse pos while dragging
        self.last_drag_pos = None
        # must be last statement
        self.init = True

    def _send_event(self, event_name, **kwargs):
        """
        :type event_name: String
        :param event_name: Name of the event
        
        Internal method. Do not call from outside.
        
        Call all event listener methods that are listening to ``event_name``
        and call the own handler method for this event 
        ``self.handle_[evnet_name]`` (if exists). Any keyword arguments will
        be passed to the listeners.
        
        See also the :ref:`event documentation <events>`\ .
        
        .. todo:: 
          Call own event handlers first?
        """
        if not self.init:
            return
        if event_name in self.listeners:
            if self.listeners[event_name]:
                for listener in self.listeners[event_name]:
                    listener(widget=self, event_name=event_name, **kwargs)
        own_handler = getattr(self, "handle_%s" % event_name, None)
        if own_handler:
            own_handler(**kwargs)
            
            
    # add and remove event handlers

    def listen(self, event_name, callback):
        """
        :type event_name: String
        :param event_name: 
            Name of an event this widget raises
        :type callback: callable
        :param callback:
            Method or function that should accept any keyword arguments 
            (\ ``**kwargs``\ ).
            
        Add an event listener. See the :ref:`event documentation <events>`\ .
        """
        if event_name not in self.listeners:
            self.listeners[event_name] = [callback]
        else:
            self.listeners[event_name].append(callback)

    def unlisten(self, event_name, callback):
        """
        :type event_name: String
        :param event_name: 
            Name of the event the handler is listening to.
        :type callback: callable
        :param callback:
            The registered event handler.
            
        Remove a previously registered event handler. 
        For more see the :ref:`event documentation <events>`\ .
        """
        if not event_name in self.listeners:
            raise IndexError, "Event not found: '%s'" % event_name
        if callback not in self.listeners[event_name]:
            raise IndexError, "Event listener not found: '%s'" % callback
        index = self.listeners[event_name].index(callback)
        del self.listeners[event_name][index]


    # get, set and change the position
    
    def _update_local_pos(self):
        old, oldr = self.local_pos, self.rect
        self.local_pos = []
        i = 0
        for value, unit, extra in self._pos:
            if unit == "%":
                value = round(self.parent.get_size(self)[i] * value / 100.) 
                value += extra
            self.local_pos.append(value)
            i += 1
        if self.anchor != "topleft":
            rect = pygame.Rect((0, 0), self.size)
            setattr(rect, self.anchor, self.local_pos)
            self.local_pos = list(rect.topleft)
        if self.local_pos != old:
               # print "changed positions:",  self.local_pos, old, self.__class__.__name__
                self.drects.append(oldr)
                self.drects.append(self.rect)

            
    def _get_pos(self):
        """
        Position of the widget.
        When setting, it is a pair of x and y coords relative to the parent
        layers topleft corner. Either absolute pixel values (eg. ``"30px"``)
        or relative to the parent size (eg. ``"50%"``). When getting, 
        it will be the absolute position on the pygame screen.
        
        For more see the :ref:`position documentation <position>`\ .
        """
        if self.update_pos_cache:
            self.update_pos_cache = False
            self._update_local_pos()
            pos = list(self.parent.get_pos(self))
            pos[0] += self.local_pos[0]
            pos[1] += self.local_pos[1]
#            i = 0
#            for value, unit in self._pos:
#                if unit == "%":
#                    value = round(self.parent.get_size(self)[i] * value / 100.)
#                pos[i] += value
#                i += 1
#            if self.anchor != "topleft":
#                rect = pygame.Rect((0, 0), self.size)
#                setattr(rect, self.anchor, pos)
#                pos = list(rect.topleft)
            if self._cached_pos != pos:
#                print "changed positions:", self._cached_pos, pos, self.__class__.__name__
#                self.drects.append(self.rect)
                self._cached_pos = pos
#                self.drects.append(self.rect)
                #!
                # .. todo::
                #
                #    need that here?
                #self.mark_dirty()
            
        return self._cached_pos

    def _set_pos(self, pos):
        """
        Position setter for the pos property. See :meth:`_get_pos`
        for more info.
        """
        _old_pos = self._pos
        old_pos = self.pos
        old_rect = self.rect
        self._pos = self._parse_setter_values(pos, "position")
        self.update_pos_cache = True
        if (self.validate_pos and
            not self.parent.get_rect(self).contains(self.rect)):
            self._pos = _old_pos
        else:
            # the old pos needs to be updated (if visible)
        #! done on pos cache update and only if localpos changed
        #    if self.visible:
        #        self.drects += [old_rect, self.rect]
            #!
            # .. todo::
            #
            #    need that here?
            #self.mark_dirty()
            #self.parent.mark_dirty()
            self._send_event("moved", changed_x=self.pos[0]-old_pos[0],
                             changed_y=self.pos[1]-old_pos[1])
                             
    pos = property(_get_pos, _set_pos)
    
    def set_abs_pos(self, pos):
        """
        :type pos: Tuple
        :param pos: 
            Absolute pixel position of the topleft widget corner
            with the topleft corner of the screen as point of origin.
        
        Set the widgets topleft position. This won't change the position unit.
        If the position is not reachable, the nearest possible position will
        be used.
        """
        rect = self.rect
        prect = self.parent.get_rect(self)
        width, height = self.size
        rect.topleft = pos
        if self.validate_pos:
            rect.left = min(
                                max(rect.left, prect.left),
                                prect.right - width,
                                )
            rect.top = min(
                                max(rect.top, prect.top),
                                prect.bottom - height,
                                )
        pos = getattr(rect, self.anchor)
        rel_pos = (pos[0] - self.parent.get_pos(self)[0],
                   pos[1] - self.parent.get_pos(self)[1])
        _pos = []
        for (old_value, old_unit, old_extra), new_value in zip(self._pos, rel_pos):

            if old_unit == "%":
                new_value = (new_value * 100. / 
                             float(self.parent.get_size(self)[len(_pos)]))
            _pos.append((new_value, old_unit, old_extra))
        _old_pos = self._pos
        old_pos = self.pos
        old_rect = self.rect
        self._pos = _pos
        self.update_pos_cache = True
        if (self.validate_pos and 
             not self.parent.get_rect(self).contains(self.rect)):
            self._pos = _old_pos
        elif self._pos != _old_pos:
            # the old pos needs to be updated (if visible)
        #    if self.visible:
        #        self.drects += [old_rect, self.rect]
                #print "new drect"
            #!
            # .. todo::
            #
            #    need that here?
            #self.mark_dirty()
            #self.parent.mark_dirty()
            self._update_local_pos()
            self._send_event("moved", changed_x=self.pos[0]-old_pos[0],
                             changed_y=self.pos[1]-old_pos[1])
    
    def move(self, offset):
        """
        :type offset: Tuple
        :param offset:
            Offset between old and new position. This can have the same
            format like a normal position (eg. ``"40px"``\ , ``"20%"``\ , 
            ``80``\ ), but the values can be positive or negative.
            
        Move the widget. This will not change the units of the current 
        position.
        """
        _offset = self._parse_setter_values(offset, "position", validate=False)
        pos = self.rect.topleft
        new_pos = []
        for value, unit, extra in _offset:
            if unit == "%":
                value *= float(self.parent.get_size(self)[i])
                value /= .100
                value += extra
            new_pos.append(value+pos[len(new_pos)])
        self.set_abs_pos(new_pos)
        
        
    # get, set and change the size

    def _get_size(self):
        """
        Size of the widget.
        When setting, it is a pair of width and height. Either absolute pixel
        values (eg ``"30px"``\ ) or relative to the parent size (eg. 
        ``"50%"``\ ). When getting, it will be the absolute width and hight.
        
        For more see the :ref:`size documentation <size>`\ .
        """
        if self.update_size_cache or self.parent.update_size_cache:
            my_x, my_y = self._get_abs_size(self._size)
            min_size = self._parse_setter_values(self.min_size, "min_size",
                                                 validate=False)
            min_x, min_y = self._get_abs_size(min_size)
            max_size = self._parse_setter_values(self.max_size, "max_size",
                                                 validate=False)
            max_x, max_y = self._get_abs_size(max_size)
            self._cached_size = [min(max_x, max(min_x, my_x)),
                                 min(max_y, max(min_y, my_y))]
            self.update_size_cache = False
        return self._cached_size
        

    def _set_size(self, size):
        """
        Size setter for the size property. See :meth:`_get_size` for more info.
        """
        new_size = self._parse_setter_values(size, "size")
        new_x, new_y = self._get_abs_size(new_size)
        my_x, my_y = self.size
        #! need resize here? (->for validate_pos?)
#        self.resize((new_x - my_x, new_y - my_y))
        #self._size = self._align_values(new_size, self._size)
        self._size = new_size
        self.update_size_cache = True
        
    size = property(_get_size, _set_size)
    
    def _get_abs_size(self, size):
        """
        :type size: tuple
        :param size: 
            A size that may be absolute or relative. Must have the
            format ``((value, unit, extra), (value, unit, , extra))`` with unit 
            as one of ``"px"`` or ``"%"`` and extra as positive or negative 
            amount of extra pixels.
        :rtype: Tuple
        :return: 
            Width and height as absolute pixel values.

        For internal use only.
        """
        retval = []
        for value, unit, extra in size:
            if unit == "%":
                # round, cause blit doesn't (blit 2.9 blits at 2)
                value = round(self.parent.get_size(self)[len(retval)] * 
                              value / 100.) + extra
            retval.append(value)
        return retval

    def resize(self, amount, anchor=None):
        """
        :param anchor:
            anchor for the resize. Eg set it to ``"topright"`` if you want to 
            move the bottomright corner.

        Resize the widget. If amount is not possible, it will resize as far as
        possible.
        """
        # amount: pair of absolute (pixel) values
        if not anchor:
            anchor = self.anchor
        x_amount, y_amount = amount
        width, height = self.rect.size
        anchor_point = getattr(self.rect, anchor)
        rect = self.rect
        old_rect = self.rect
        min_width, min_height = self._get_abs_size(
                        self._parse_setter_values(self.min_size, "min_size"))
        max_width, max_height = self._get_abs_size(
                        self._parse_setter_values(self.max_size, "max_size"))
        while min_width <= width + x_amount <= max_width:
            rect.width = width + x_amount
            setattr(rect, anchor, anchor_point)
            if (not self.validate_pos or
                self.parent.get_rect(self).contains(rect) or x_amount == 0):
                break
            x_amount += 1 if x_amount < 0 else -1
        
        while min_height <= height + y_amount <= max_height:
            rect.height = height + y_amount
            setattr(rect, anchor, anchor_point)
            if (not self.validate_pos or 
                self.parent.get_rect(self).contains(rect) or y_amount == 0):
                break
            y_amount += 1 if y_amount < 0 else -1
        if rect.size != self.rect.size:
            # keep old units
            self._size = self._align_values(self._size, 
                                    ((rect.width, "px", 0), (rect.height, "px", 0)))
            self.update_size_cache = True
            self.set_abs_pos(rect.topleft)
            if not self.rect.contains(old_rect) and self.visible:
                self.drects += [old_rect, self.rect]
            self._send_event("resized", changed_w=x_amount,
                         changed_h=y_amount)
        
        
    def set_size(self, new_size, anchor=None):
        """
        Not yet implemented.
        
        .. todo::
        
            Implement :meth:`Widget.set_size`
            
        """
        # like set_abs_pos for size
        raise Exception, "Not yet implemented!"


    # misc methods regarding position and size

    def _get_rect(self):
        """
        A rect containing this widget. This is a read-only property, build from
        :attr:`size` and :attr:`pos`\ . 
        To change the position of the widget use its :attr:`pos` (and 
        :attr:`size` and :attr:`anchor`\ ) attributes. 
        Changing the rect will not affect the widget position.
        """
        return pygame.Rect((self.pos + self.size))
    
    rect = property(_get_rect)
        
        
    def _align_values(self, value1, value2, include_units=True):
        """
        :type value1: Tuple
        :param value1:
            Internal position or size like ``((value, unit), (value, unit))`` 
            with unit as one of  ``"px"`` or ``"%"``\ .
        :type value2: Tuple
        :param value2:
            Same as `value1`
        :type include_units: boolean
        :param include_units: 
            If False, the return value will be a tuple of two single
            values without units.
            
        Convert the units of `value2` to these of `value1`.
        """
        new_value2 = []
        for (amount1, unit1, extra1), (amount2, unit2, extra2) in zip(value1, value2):
            if not unit1 == unit2:                
                if unit1 == "%" and unit2 == "px":
                    amount2 = ((amount2-extra1) * 100. / 
                            float(self.parent.get_size(self)[len(new_value2)]))
                elif unit1 == "px" and unit2 == "%":
                    amount2 = (amount2 * 
                            float(self.parent.get_size(self)[len(new_value2)])
                            / 100.) + extra1
                else:
                    raise RuntimeError, "Invalid units found"
            new_value2.append((amount2, unit1, extra1) if include_units else amount2)
        return new_value2


    def _parse_setter_values(self, values, name, validate=True):
        """
        :type values: Tuple
        :param values: 
            A :ref:`position` or :ref:`size` value.
        :type name: String
        :param name:
            Name of the value to convert (eg. "pos", "size", "max_size", ...).
            This will only be used in Exceptions.
        :type validate: boolean
        :param validate:
            Whether the values should be validated or not.
            
        Convert user values for positions or sizes to the internal format.
        """
        new_values = []
        if not len(values) == 2:
            raise ValueError, "%s argument must be a (x, y) tuple" % name
        for p in values:
            unit = "px"
            extra = 0
            if type(p) == str:
                if "%" in p:
                    unit = "%"
                    p, e = p.split("%", 1)
                    if e:
                        extra = int(e)
                elif p.endswith("px"):
                    p = p[:-2]
            try:
                value = int(p)
            except ValueError:
                raise (ValueError,
                       "%s must consist of pixel or percentage values" % name)
            if validate and self.validate_pos:
                if unit == "%":
                    if not (0 <= value <= 100):
                        raise ValueError, "%s out of range: %s%%" % (name, p)
                elif not 0 <= value <= self.parent.get_size(self)[
                                                        len(new_values)]:
                    raise ValueError, (
                                   "%s out of range: %spx (parent size: %s): "
                                   % (name, p, self.parent.get_size(self)))
            new_values.append((value, unit, extra))
        return new_values


    def update(self, events):
        """
        :type events: list
        :param events: pygame events that may be used
        :rtype: list
        :return: pygame events that were not used
        
        Process events and update the widget.
        """                        
        orig_events = events[:]
        self.update_pos_cache = True                
        self.update_size_cache = True
        if self.visible:
            events = self.handle_events(events)
            events = self._handle_default_events(events)
            self.update_widget()

            styles = self.get_styles()
            if tuple(self.size) != self.image.get_size():
                self.mark_dirty()
                #! use style("image") with transparent image as default
                self.image = pygame.Surface(self.size, loc.SRCALPHA)
            styles_hash = dict([(k, hash(v)) for k, v in styles.iteritems()])
            if styles_hash != self._cached_styles or self.dirty:
                self.image.fill((0, 0, 0, 0,))
                self.blit_widget(styles)
                self._cached_styles = styles_hash
                self.dirty = False
                self.drects.append(self.rect)
                
        drects = self.drects
        self.drects = []    
        if self.pass_events:
            events = orig_events    
        return events, drects

        

    def update_widget(self):
        """
        Hook, called once per frame.
        """
        pass

    def blit_widget(self, styles):
        """
        :type styles: dict
        :param styles:
            Style attributes used to blit the widget.
        :rtype: list
        :returns:
            List of dirty rects
        reblit self.image. You should use attributes from ``styles`` only.
        """
        if "bgcolor" in styles and styles["bgcolor"]:
            self.image.fill(styles["bgcolor"])
            return [self.rect]
        if "image" in styles and styles["image"]:
            image = styles["image"]
            rect = self.image.get_rect()
            image_rect = image.get_rect(center=rect.center)
            self.image.blit(image, image_rect)
            return [self.rect]

        
    def debug(self, screen):
        pass
        

    def blit(self, screen):
        """
        :type screen: ``pygame.Surface``
        :param screen:
            Surface to blit the widget onto.
        :rtype: list
        :returns:
            A list that contains either a Rect if the Widget was changed or
            None. Can be passed to ``pygame.display.update``\ .

        Blit the widget onto a surface. Normally this will be called by
        the parent Widget only.
        """
        if not self.visible:
            return
        # cut of parts outside the parents inner rect
        if self.parent:
            rect = self.parent.get_rect(self).clip(self.rect)
        else:
            rect = self.rect
        # rect will have a size of (0, 0) when it is not visible
        # (outside the parent rect)
        if rect:
            screen.blit(self.image, self.localize_rect(rect), 
                    
                    self.localize_rect(rect, self.rect))

        
    def localize_rect(self, rect, rel=None):
        """
        :type rect: ``pygame.Rect``
        :param rect:
            Rect to localize.
        :type rel: ``pygame.Rect`` or None
        :param rel:
            Rect to use as origin. Defaults to ``self.parent.rect``
        :rtype: ``pygame.Rect``
        :returns:
            Localized rect.
        Turns a global rect into a local one (ie. relative ``rel``).
        """
        if not (self.parent or rel):
            return rect
        # no get_rect since its absolute to parent surface topleft (for blitting)
        rel = rel or self.parent.rect
        return rect.move(-rel.left, -rel.top)
        

    def _handle_default_events(self, events):
        """
        :type events: list
        :param events: pygame events that may be used
        :rtype: list
        :return: pygame events that were not used

        Internal method to wrap some mouse events to own events. 
        (See :ref:`events`\ ).
        """
        wrap_map = {loc.MOUSEBUTTONDOWN: "mouse_down",
                    loc.MOUSEBUTTONUP: "mouse_up",
                    loc.MOUSEMOTION: "mouse_moved"}
        mouse_pos = pygame.mouse.get_pos()
        old_dragging = self.dragging
        self.dragging = [a and b for a, b in
                         zip(self.dragging, pygame.mouse.get_pressed())]
        if any(old_dragging) and not any(self.dragging):
            self._send_event("mouse_stopped_dragging", pos=mouse_pos)
        if any(self.dragging):
            rect = self.rect
            # dragged rect
            rect.topleft = (mouse_pos[0]-self.dragging_anchor[0],
                            mouse_pos[1]-self.dragging_anchor[1])
            self._send_event("mouse_dragged", buttons=self.dragging,
                                  rel=(rect.left-self.rect.left,
                                       rect.top-self.rect.top,),
                                  abs=(mouse_pos[0]-self.last_drag_pos[0],
                                       mouse_pos[1]-self.last_drag_pos[1]))
            self.last_drag_pos = mouse_pos
        else:
            self.dragging_anchor = None

        mouse_over = self in self.get_gui().active_widgets_under_mouse
        
        # update mouseover flag first and then send mouse_left event
        if not mouse_over and self.mouse_over:
                self.mouse_over = False
                self._send_event("mouse_left", last_pos=self.last_mouse_pos)

        if mouse_over:
            if not self.mouse_over:
                self.mouse_over = True
                self._send_event("mouse_entered", pos=mouse_pos)
                

            self.last_mouse_pos = mouse_pos
            not_used = []
            for event in events:
                if event.type in wrap_map:
                    self._send_event(wrap_map[event.type], **event.dict)
                    if event.type == loc.MOUSEBUTTONDOWN and event.button < 4:
                        self.dragging[event.button-1] = True
                        my_x, my_y = self.pos
                        self.dragging_anchor = [mouse_pos[0]-my_x, 
                                                mouse_pos[1]-my_y]
                        self.last_drag_pos = mouse_pos
                        self.get_gui().set_focus(self)
                        #! 
                        # .. todo::
                        #   remove quick fix
                        self._send_event("post_"+wrap_map[event.type], **event.dict)
                else:
                    not_used.append(event)
            return not_used
        return events
        

    def collide_point(self, point):
        """
        :type point: Tuple
        :param point:
            A point (x, y) on the pygame screen.
        :rtype: boolean
        :return:
            Whether the point touches the widget. By default, this simply
            uses ``self.rect.collidepoint``\ .
            
        Overwrite this method to get more accurate mouse events 
        for non-rectangular widgets.
        """
        return self.visible and self.rect.collidepoint(point)

    def handle_events(self, events):
        """
        :type events: list
        :param events: pygame events that may be used
        :rtype: list
        :return: pygame events that were not used

        Overwrite this method to process raw pygame events.
        """
        return events

    
    def get_styles(self):
        """
        :rtype: dict
        :return:
            all styles and attributes used to blit the widget and that 
            may change
            
        return value will be passed to :meth:`~Widget.update_widget`.
    
        By default, this returns a dict with all styles with stylenames in
        :attr:`used_syles`
        """
        styles = dict([(name, self.style(name, self)) 
                        for name in self.used_styles])
        return styles
    
    
    def get_gui(self):
        """
        :rtype: :class:`GUI`
        :return: the main :class:`GUI` instance that contains this widget.
        
        Get the topmost :class:`GUI` widget.
        """
        return self.parent.get_gui()
    
    def is_ancestor(self, widget):
        """
        :type widget: :class:`~gui.base.Widget`
        :param widget: a widget that could be an ancestor
        :rtype: boolean
        :return: if `widget` is an ancestor
        
        Test for direct or indirect relation.
        """
        parent = self.parent
        while parent:
            if parent == widget:
                return True
            parent = parent.parent
        return False    
    
    
#    def get_widget(self, point):
#        """
#        :type point: Tuple
#        :param point:
#            A point (x, y) on the pygame screen.
#        :rtype: :class:`Widget` or ``False``
#        :return: Topmost widget at this point.
        
#        Get the topmost widget at a certain point.
#        """
#        return (not self.pass_events and self.collide_point(point) and 
#                self.visible and self.rect.clip(
#                        self.parent.get_rect(self)).collidepoint(point) and
#                self or False)
                
    def get_suffix(self, style_name):
        """
        :type style_name: String
        :param style_name:
            Name of a requested style attribute.
        :rtype: String
        :return:
            Style name suffix.
        
        Return a string that will be added to style_name before searching
        for the style attribute. See docs on :ref:`styling`\ .
        
        .. todo::
           test for performance issures
        """
        if self.highlight and (any(self.dragging) or self.mouse_over):
            return "active"
        if self.focused:
            return "focused"
            
    def mark_dirty(self):
        """
        Mark Widget as dirty.
        """
        self.dirty = True

    def show(self):
        """
        Show the widget if it is hidden.
        """
        if not self.visible:
            self._send_event("shown")
            self.visible = True
            self.drects.append(self.rect)
            self.mark_dirty()

    def hide(self):
        """
        Hide the widget if it is visible.
        """
        if self.visible:
            self._send_event("hidden")
            self.visible = False
            self.drects.append(self.rect)
            self.mark_dirty()

    def toggle(self):
        """
        Toggle visibility.
        """
        if self.visible:
            self.hide()
        else:
            self.show()
            
    def delete(self):
        self.parent.remove(self)

#    def __nonzero__(self):
#        """
#        :rtype: boolean
#        :return: whether the widget is visible or not.
        
        
#        .. todo::
        
#            is this really useful?
            
#        """
#        return self.visible 

class Layer(Widget):
    """
    A :class:`Layer` is a :class:`Widget` that can contain other widgets
    (and Layers, since Layer is a child class of Widget). You can use Layers
    either as a frame, ie. to gorup different widgets, or to build own 
    *superwidgets*, ie. widgets that contain other widgets. Child
    widgets cannot overlap a Layer, they must be completely contained.
    """
    focusable = False
    def __init__(self, parent, size=None, pos=(0, 0), **kwargs):
        """
        The arguments are the same like for :class:`Widget`\ . Only that `size`
        will be set to the size of the parent widget and `pos` to ``(0, 0)``
        by default.
        """
        if not size:
            size = parent.get_size(self)
        Widget.__init__(self, parent, size, pos, **kwargs)
        self.widgets = []
        
    def add(self, widget, index="end", send_event=True):
        """
        :type widget: :class:`Widget`
        :param widget:
            A Widget to add to the Layer.
        :type index: int or ``"end"``
        :param index:            
            Index of the new Widget in the Widget list. The rearmost position 
            is at index 0, the topmost position at ``"end"`` (default).
        :type send_event: boolean
        :param send_event:
            Whether a ``"widget_added"`` event should be send. Default is True.

        Add a Widget at the given index to the Layer.
        """
        if index == "end":
            self.widgets.append(widget)
        else:
            self.widgets.insert(index, widget)
        if send_event:
            self._send_event("widget_added", new_widget=widget, index=index)
            
    def remove(self, widget, send_event=True):
        """
        :type widget: :class:`Widget`
        :param widget:
            A Widget to remove from the Layer.
        :type send_event: boolean
        :param send_event:
            Whether a ``"widget_removed"`` event should be send.
            Default is True.

        Remove a Widget from the Layer.
        """
        del self.widgets[self.widgets.index(widget)]
        self.dirty = True
        if send_event:
            self._send_event("widget_removed", old_widget=widget)
            
    def get_index(self, widget):
        """
        :type widget: :class:`Widget`
        :param widget:
            A child Widget of the Layer.
        :rtype: int
        :return: 
            Index of the given Widget.
        """
        if widget.parent != self:
            raise ValueError, "Widget must be child widget"
        
        return self.widgets.index(widget)
        
    def change_index(self, widget, new_index="end"):
        """
        :type widget: :class:`Widget`
        :param widget:
            A child Widget of the Layer.
        :type new_index: int or ``"end"``
        :param new_index:            
            The new index for the Widget.

        Change the index of a child Widget.
        """
        self.remove(widget, send_event=False)
        self.add(widget, new_index, send_event=False)
        self.mark_dirty()
        self._send_event("widget_index_changed", changed_widget=widget,
                         new_index=new_index)
                         
    def rel_chage_index(self, widget, amount):
        """
        :type widget: :class:`Widget`
        :param widget:
            A child Widget of the Layer.
        :type amount: int
        :param amount:            
            Distance form the current index. Positive and negative values
            are possible. The new index will be current index + amount.

        Change the index of a child Widget relatively to its current index.
        """
        old_index = self.get_index(widget)
        # don't create negative indices
        new_index = max(old_index + amount, 0)
        self.change_index(widget, new_index)
        
    def widget_up(self, widget):
        """
        Shortcut for :meth:`~Layer.rel_change_index`\ . Moves a child
        widget one position forward.
        """
        self.rel_chage_index(widget, 1)
        
    def widget_down(self, widget):
        """
        Shortcut for :meth:`~Layer.rel_change_index`\ . Moves a child
        widget one position backward.
        """
        self.rel_chage_index(widget, -1)

    def update(self, events):
        """
        :type events: list
        :param events: 
            Pygame events since the last frame that were not yet used
        :returns: list of events that were not used and 
                  can be passed to the next widget
                  
        Update all children, from foreground to background,
        then the Layer itself.
        """
        if self.visible:
            self.update_pos_cache = True
            self.update_size_cache = True
            child_drects = []
            for widget in reversed(self.widgets):
                events, new_drects = widget.update(events)
                child_drects += new_drects
            if self.pass_events:
                # don't save the returned events if event transparent
                self.handle_events(events)
                self._handle_default_events(events)
            else:
                events = self.handle_events(events)
                events = self._handle_default_events(events)
                
            
            self.update_widget()
            
            styles = self.get_styles()
            if tuple(self.size) != self.image.get_size():
                self.mark_dirty()
            
                self.image = pygame.Surface(self.size, loc.SRCALPHA)
            styles_hash = dict([(k, hash(v)) for k, v in styles.iteritems()])
            if styles_hash != self._cached_styles:
                self.mark_dirty()
            if self.dirty or child_drects:  
                if self.dirty:
                    self.drects.append(self.rect)
                else:
                    self.drects += child_drects
                self.image.fill((0, 0, 0, 0,))
                self.blit_widget(styles)
                for widget in self.widgets:
                    widget.blit(self.image)
                self.blit_widget_top(styles)
                self._cached_styles = styles_hash
                self.dirty = False
        drects = self.drects 
        self.drects = []
        return events, drects
    
    def blit_widget_top(self, styles):
        """
        called after all widgets were blitted
        """

    def debug(self, screen):
        for widget in self.widgets:
            widget.debug(screen)
    
#    def update_cache(self):
#        """
#        update cached values and styles.
#        """
#        for widget in self.widgets:
#            widget.update_cache()


#    def blit(self, screen):
#        """
#        :type screen: pygame.Surface
#        :param screen:
#            Surface on which to blit.
#        :rtype: list
#        :return: Rects that need to be updated (dirty rects).
        
#        Blit the Layer and all its child Widgets,
#        from background to foreground.
        
#        """
#        if not self.visible:
#            return [None]
#        drect = Widget.blit(self, screen)
#        child_drects = []
    
#        # don't need clipping here anymore, since widget hanldes this now.
#        #orig_clip = screen.get_clip()
#        #screen.set_clip(orig_clip.clip(self.get_rect(self)))
#        for widget in self.widgets:
#            child_drects += widget.blit(self.image)
#        #screen.set_clip(orig_clip)
    
#    # if the Layer itself needs to be updated, discard child drects since
#    # they are all inside the own Rect
#        return drect or child_drects

#    def blit_widget(self, styles):
#        """
#        :type screen: pygame.Surface
#        :param screen:
#            Surface on which to blit.
#        :rtype: list
#        :return: Rects that need to be updated (dirty rects).
        
#        Blit the Layer and all its child Widgets,
#        from background to foreground.
        
#        """
#        if not self.visible:
#            return [None]
#        drect = [None]
#        if styles == self._cached_styles:
#            drect = Widget.blit_widget(self, styles)    
#     #   child_drects = []
    
#        # don't need clipping here anymore, since widget hanldes this now.
#        #orig_clip = screen.get_clip()
#        #screen.set_clip(orig_clip.clip(self.get_rect(self)))
#    #    for widget in self.widgets:
#    #        child_drects += widget.blit(self.image)
#        #screen.set_clip(orig_clip)
    
#    # if the Layer itself needs to be updated, discard child drects since
#    # they are all inside the own Rect
#        print drect != [None] and drect or child_drects
#        return drect != [None] and drect or child_drects

    def get_widgets(self, point):
        """
        :type point: Tuple
        :param point:
            A point (x, y) on the pygame screen.
        :rtype: list
        :return: Widgets at ``point``, topmost at the end
        
        """
        if not self.collide_point(point):
            return []
        widgets = [self]
        for widget in (self.widgets):
            if not self.get_rect(widget).collidepoint(point):
                continue
            if isinstance(widget, Layer):
                widgets += widget.get_widgets(point)
            elif widget.collide_point(point):
                widgets.append(widget)
        return widgets
    

    def get_rect(self, widget):
        """
        :type widget: :class:`Widget`
        :param widget:
            A child widget.
        :rtype: ``pygame.Rect``
        :return: 
            The :ref:`inner rect` for the widget.


            
        .. todo::
        
            document "borderwidth" standard syle attribute
        """
        #return self.rect
        bd = self.style("borderwidth", self)
        my_rect = self.rect
        if bd:
            if type(bd) == int:
                my_rect.inflate_ip(2*-bd, 2*-bd)
            else:
                my_rect.inflate_ip(2*-bd[0], 2*-bd[1])
            my_rect.center = self.rect.center
        return my_rect
    

    def get_size(self, widget):
        """
        get size for a certain widget
        """
        return self.get_rect(widget).size
    
    def get_pos(self, widget):
        """
        pos size for a certain widget
        """
        return self.get_rect(widget).topleft
    
    
    
class GUI(Layer):
    """
    GUI is the topmost :class:`Layer` and the only :class`Widget` that
    has no parent. Thus, you need a :class:`GUI` before you can create other
    Widgets.
    """
    # overwrite the `pos` and `size` properties of Widget
    size = None
    pos = None
    def __init__(self, size, pos=(0, 0), style={}, pass_events=True):
        """
        :type size: Tuple
        :param size:
            Absolute size of the GUI.
        :type pos: Tuple
        :param pos:
            Absolute position of the GUI on the pygame screen.
            Defaults to ``(0, 0)``.
            
        .. todo::
        
          should `pass_events` default to True?

        """
        if isinstance(style, dict):
            # create style without fallback
            style = Style(**style)
        Layer.__init__(self, None, size, pos, style=style, 
                       pass_events=pass_events)
        self.focused_widget = None
        # widgets under the mouse pointer, last one is the topmost
        self.widgets_under_mouse = []
        # widgets that should have mouse_over. Normally, this is just 
        # self.widgets_under_mouse[-1], but if this is a "event transparent"
        # widget, the next one is added and so on
        self.active_widgets_under_mouse = []

        
    def handle_events(self, events):
        """
        :tpye events: list
        :param events:
            Pygame events to process.
        :return:
            List of pygame events that were not used.
            
        Resize the GUI on ``pygame.VIDEORESIZE``.
        
        .. todo::
        
            Make this optional. This ignores the current :attr:`~Widget.size`
            and :attr:`~Widget.pos`\ .

        """
        #! make this optional
        for event in events:
            if event.type == loc.VIDEORESIZE:
                self.size = event.size
        return events
    
    def get_gui(self):
        """
        :rtype: :class:`GUI`
        :return: self
        
        Hook for child Widgets to get the topmost GUI instance.
        """
        return self
    

    
#    def get_widget(self, point):
#        """
#        :type point: Tuple
#        :param point:
#            A point (x, y) on the pygame screen.
#        :rtype: :class:`Widget`
#        :return: Topmost widget at this point.
        
#        Get the topmost widget at a certain point.
#        This will always return a widget.
#        """
#        for widget in reversed(self.widgets):
#            w = widget.get_widget(point)
#            if w:
#                return w
#        return self

    def set_focus(self, widget):
        """
        :type widget: :class:`Widget` or `None`
        :param widget:
            Widget that was clicked or otherwise selected.
            
        Set the :ref:`focus <focus>` to a new Widget or to None. 
        If `Widget` is not focusable or None, the :ref:`focus  <focus>` will 
        be set to None.
        """
        if self.focused_widget == widget:
            return
        if self.focused_widget:
            self.focused_widget.focused = False
            self.focused_widget._send_event("focus_lost")
        old_focused = self.focused_widget
        if widget and widget.focusable:
            self.focused_widget = widget
            widget.focused = True
            self._send_event("new_focus", old_focused=old_focused,
                              new_focused=self.focused_widget)
            self.focused_widget._send_event("focused")
        else:
            self.focused_widget = None
            self._send_event("focus_lost", old_focused=old_focused,)
            
    def update_widget(self):
        self.widgets_under_mouse = self.get_widgets(pygame.mouse.get_pos())
        self.active_widgets_under_mouse = []
        for widget in reversed(self.widgets_under_mouse):
            self.active_widgets_under_mouse.append(widget)
            if not widget.pass_events:
                break
        