# coding: utf-8
"""
GUI image handlers - provide images in any size
"""
import os

import pygame
import pygame.locals as loc

class ImageHandler(object):
    
    #: Cache dict to save already loaded images.
    #:_cache = {image_load_hash:{"original":Image, size:Image}
    _cache = {}
    def __init__(self, path, extra=None):
        """
        :type path: String
        :param path:
            Absolute or relative path to an image file.
        :type extra: Tuple or String or None
        :param extra:
            * If it is a color (3-Tuple), it will be used as colorkey.
            * If it is a position (2-Tuple), the pixel at this position will
            be used as colorkey.
            * If it is ``"alpha"`` the image will be converted to use per-pixel
            alpha.
            
        """
        self.hash = None
        self.last_sizes = {}
        self.set_image(path, extra)
    
        
        
    def set_image(self, path, extra=None):
        """
        Same attributes as meth:`__init__`\ .
        Set a new image.
        """
        path = os.path.abspath(path)
        try:
            image = pygame.image.load(path)
        except pygame.error, message:
            raise SystemExit, message
        if extra:
            if len(extra) == 2:
                extra = image.get_at(extra)[:3]
            if len(extra) == 3:
                #image = image.convert()
                image.set_colorkey(extra, loc.RLEACCEL) #! use loc.RLEACCEL here?
            elif extra == "alpha":
                image = image.convert_alpha()
        self.hash = hash((path, extra))
        if hash not in self._cache:
            self._cache[self.hash] = {"original":image}
            #self.last_size = image.get_size()
            
        
    def get_original(self):
        return self._cache[self.hash]["original"]
    
    def set_original(self, image):
        """
        Set original image. Allows you to modify loaded images manually.
        This will affect all styles using this image.
        
        """
        self._cache[self.hash]["original"] = image
        # clear cache
        for k in self._cache[self.hash]:
            if k != "original":
                del self._cache[self.hash][k]
        

        
    def __call__(self, widget, size=None):
        """
        :type widget: :class:`Widget`
        :param widget:
            The calling Widget.
        :type size: Tuple
        :param rect:
            Size of the new image.
            
        Return the image for a Widget. If size is ``None``\ , 
        ``widget.rect.size`` will be used.
        """
        new_size = size or widget.rect.size
        last_size = self.last_sizes.get(widget, (0, 0))
        if new_size != last_size or new_size not in self._cache[self.hash]:
            if last_size in self._cache[self.hash]:
                del self._cache[self.hash][last_size]
            self.last_sizes[widget] = new_size
            if not new_size in self._cache[self.hash]:
                image = self.build(new_size)
                self._cache[self.hash][new_size] = image
                return image
        return self._cache[self.hash][new_size]
        
        
    def build(self, size):
        """
        :type size: Tuple
        :param size:
            Pair of (width, height) in pixels.
        :rtype: ``pygame.Surface``
        :return:
            New version of the image resized to `size`.
            
        Custom ImageHandlers should overwrite this method.
        """
        return self._cache[self.hash]["original"].subsurface((0, 0) + size)
    
class ScaleImageHandler(ImageHandler):
    """
    ImageHandler that uses ``pygame.transform.scale`` for resizing
    """
    def build(self, size):
        return pygame.transform.scale(self._cache[self.hash]["original"],
                                             size)


class SmoothScaleImageHandler(ImageHandler):
    """
    ImageHandler that uses ``pygame.transform.smoothscale`` for resizing
    """
    def build(self, size):
        return pygame.transform.smoothscale(self._cache[self.hash]["original"],
                                             size)
                                             
                                             
class FrameImageHandler(ImageHandler):
    def set_image(self, path, extra=None):
        ImageHandler.set_image(self, path, extra)
        original = self.get_original()
        w, h = original.get_size()
        half_x = int(round(w/2.))
        half_y = int(round(h/2.))

        parts = []
        widths = ((0,half_x),(half_x, half_x+1), (half_x+1,w))
        heights= ((0,half_y),(half_y, half_y+1), (half_y+1,h))

        for x1, x2 in widths:
            for y1, y2 in heights:
                s = pygame.surface.Surface((x2-x1, y2-y1), 0, original)
                s.blit(original, (0,0), pygame.rect.Rect((x1,y1,x2,y2)))
                parts.append(s)
        self._cache[self.hash]["parts"] = parts
        self._cache[self.hash]["size"] = original.get_size()
        
                
    def build(self, size):
        new_width, new_height = size
        parts = self._cache[self.hash]["parts"]
        orig_width, orig_height = self._cache[self.hash]["size"]
        # cannot make it smaller
        new_width = max(new_width, orig_width)
        new_height = max(new_height, orig_height)

        # new size - side sizes
        resize_width = new_width - orig_width + 1
        resize_height = new_height - orig_height + 1

        # crete surface of the new size. pass the first part to get the
        # same pixel format.
        new_image = pygame.Surface((new_width, new_height), 0, parts[0])
        x_offset, y_offset = 0, 0
        for i, part in enumerate(parts):
            if i in (1,3,4,5,7):
                new_part_width, new_part_height = part.get_size()
                if i in (1, 4, 7):
                    new_part_height = resize_height
                if i in (3, 4, 5):
                    new_part_width = resize_width
                part = pygame.transform.scale(part, 
                                              (new_part_width, new_part_height))
            new_image.blit(part, (x_offset, y_offset))
            y_offset += part.get_height()
            if not (i+1) % 3:
                y_offset = 0
                x_offset += part.get_width()
        return new_image


class XFrameImageHandler(FrameImageHandler):
    """
    Like :class:`FrameImageHandler`, but uses the frame only for horizontal
    resizing and normal scaling for vertical resizing.
    
    .. todo:: 
    
      improve and add YFrameImageHandler
      
    """
    def set_image(self, path, extra=None):
        ImageHandler.set_image(self, path, extra)
        original = self.get_original()
        w, h = original.get_size()
        half_x = int(round(w/2.))
        half_y = h-1

        parts = []
        widths = ((0,half_x),(half_x, half_x+1), (half_x+1,w))
        heights= ((0,half_y),(half_y, half_y+1), (half_y+1,h))

        for x1, x2 in widths:
            for y1, y2 in heights:
                s = pygame.surface.Surface((x2-x1, y2-y1), 0, original)
                s.blit(original, (0,0), pygame.rect.Rect((x1,y1,x2,y2)))
                parts.append(s)
        self._cache[self.hash]["parts"] = parts
        self._cache[self.hash]["size"] = original.get_size()



class PlainColorImageHandler(ImageHandler):
    """
    A :class:`ImageHandler` for plain-colored surfaces.
    Should only be used for RGBA colors. For RGB use 
    ``Surface.fill(color, rect)``
    """
    #: Cache dict to save already loaded images.
    #:_cache = {image_load_hash:{"original":Image, size:Image}
    _cache = {}
    def __init__(self, color):
        """
        :type color: Tuple
        :param color:
            Either a rbg or rgba color.
            
        """
        self.hash = None
        self.flag = None
        self.color = None
        self.last_sizes = {}
        self.set_color(color)
        

    def set_color(self, color):
        """
        Set a new color
        """        
        print "set color"
        self.hash = hash(color)
        self.color = color
        self.flag = len(self.color) == 4 and loc.SRCALPHA or 0
        self._cache[self.hash] = {}
            
        
    def get_original(self):
        return self.color
            
        
    def build(self, size):
        """
        Custom ImageHandlers should overwrite this method.
        """
        surf = pygame.Surface(size, self.flag)
        surf.fill(self.color)
        return surf