# coding: utf-8
"""
GUI Widgets
"""

import math

import pygame
import pygame.locals as loc

from base import Widget, Layer
from style import Style


class Text(Widget):
    _aligns = ("left","centered", "right")
    used_styles = ["image", "font", "textcolor", "textbackground"]
    template = "%s"
    def __init__(self, parent, size, pos, text, **kwargs):
        """
        In addition to the default :class:`~gui.base.Widget` parameters
        and events, this widget provides the following:
        
        :type size: :ref:`Size` or "auto"
        :param size:
            Size of the text. If the renered text is larger than the 
            Widget size, it will be trimmed. If set to ``"auto"``\ , the
            Widget will be resized whenever a new text is set to fit the
            new rendered text size.
        :type text: String
        :param text:
            The text to show.
        :type align: String
        :param align:
            Vertical align of the text. Must be one of *right*, *left* or 
            *centered*\ , the default.
        :type template: string
        :param template: 
            Template string to pre- and/or suffix the editable text.
            Must contain "%s" exactly once.             
            
        .. describe:: Events
        
            text_set - the text was changed
        
           
        
        """
#        self._text = ""
        self.auto_resize = size == "auto"
        if self.auto_resize:
            size = (1, 1)
        self.image = None
        self.text_rects = []
        self.align = kwargs.get("align", "centered")
        if self.align not in self._aligns:
            raise ValueError, "Invalid align value '%s'! Must be one of %s" % (
                                 self.align, ", ".join(self._aligns))
        if "align" in kwargs:
            del kwargs["align"]
        Widget.__init__(self, parent, size, pos, **kwargs)
        self.text = text
        
        self.template = kwargs.get("template", self.template)
        if self.template.count("%s") != 1:
            raise ValueError('Template must contain exactly one formatting ' +
                              'character: %s' % self.template)
                            
        if self.auto_resize:
            self.blit_widget(self.get_styles())
            
        
    
#    def _get_text(self):
#        """
#        The text to show.
#        """
#        return self._text
    
#    def _set_text(self, new_text):
#        old_text = self._text
#        self._text = new_text
#        self.update_text()
#        self._send_event("text_set", old_text=old_text, new_text=self._text)

#    text = property(_get_text, _set_text)
    
    def get_styles(self):
        styles = Widget.get_styles(self)
        styles["text"] = self.template % self.text
        styles["align"] = self.align
     #   if self.mouse_over:
     #       print styles
        return styles

    def blit_widget(self, styles):
        """
        reblit the text
        """
  #      print "blit text", self.text
        drects = Widget.blit_widget(self, styles) or [self.rect]
        lines = []
        self.text_rects = []
        max_width = 0
        height = 0
        #line_height = self.font.get_linesize()

        for line in styles["text"].splitlines():
            line = self.render_line(line, styles)
            lines.append(line)
            max_width = max(max_width, line.get_width())
            height += line.get_height()
        size = (max_width, height)
        if self.auto_resize:
            self.size = size
        else:
            size = self.rect.size
      #  self.image = pygame.Surface(size, loc.SRCALPHA)
        rect = self.image.get_rect()
        
        if styles["align"] == "centered":
            anchor = "midtop"
        else:
            anchor = "top%s" % styles["align"]
        x_pos = getattr(rect, anchor)[0]
        height = 0
        for line in lines:
            rect = line.get_rect(**{anchor:(x_pos, height)})
            self.image.blit(line, rect)
#            rect.move_ip(*self.parent.get_pos(self))
            self.text_rects.append(rect)
            height += line.get_height()#line_height
            
        drects, self.drects = self.drects, drects
        drects.append(self.drects)
        
        return drects
            
#    def update_cache(self):
#        """
#        Hook to update the text on global call.
#        """
#        self.update_text()
            
    def render_line(self, line, styles):
        """
        :type line: String
        :param line:
            One line of text to render.
        :rtype: ``pygame.Surface``
        :return:
            The rendered text.
            
        Render a line of text.
        """
        color = styles.get("textcolor", (0,0,0)) or (0, 0, 0)
        font = styles.get("font", None)
        if not font:
            font = pygame.font.Font(None, 20)
        bg = styles["textbackground"]
        if type(line) != unicode:
            line = unicode(line, "utf-8")
        if bg:
            return font.render(line, 1, color, bg)
        else:
            return font.render(line, 1, color)
            
#    def blit(self, screen):
#        """
#        Blit the text on a pygame surface ``screen``.
#        """
#        drects = Widget.blit(self, screen)
#        if self.visible:
#            drects.append(screen.blit(self.image, self.pos))
#        return drects
    
    def collide_point(self, point):
        """
        Check for collision with any line.
        """
        if not Widget.collide_point(self, point):
            return False
        px, py = point
        my_x, my_y = self.pos
        x, y = px-my_x, py-my_y
        for rect in self.text_rects:
            if rect.collidepoint((x, y)):
                return True                
        return False
    
    
    
        
#    def debug(self, screen):
#        pygame.draw.rect(screen, (200,100,100), self.rect, 1)
#        for rect in self.text_rects:
#            rect = rect.move(self.rect.topleft)
#            pygame.draw.rect(screen, (100,100, 200), rect, 1)
#        mx, my = pygame.mouse.get_pos()
#        if self.rect.collidepoint((mx,my)):
#            px, py = mx, my
#            my_x, my_y = self.pos
#            x, y = px-my_x, py-my_y
#            pygame.draw.rect(screen, (255,0,255), (x,y,10,10))

        

class TextInput(Text):
    #! todo
    # - text navigation (up/down in multiline)
    # - on mouse_down/focus set cursor to mouse pos
    multiline = True
    highlight = True
    def __init__(self, *args, **kwargs):
        """
        An editable text widget.
        Extra arguments:

        :type multiline: boolean
        :param multiline: 
            If True (default), newlines are allowed.
            
        Extra events:
        
        **pressed_return**
          User pressed return. (non-multiline only)
        """
        Text.__init__(self, *args, **kwargs)
        self.cursor_pos = 0
        if "multiline" in kwargs:
            self.multiline = kwargs["multiline"]
        self._time = 0
        self.insert = True
        self.cursor = "|"
        # blink interval in milliseconds
        self.interval = 500
        self._orig_key_repeat = pygame.key.get_repeat()
        #: store raw text (without cursor)
        self._text = self.text
    
    def handle_focused(self, **kwargs):
        self._orig_key_repeat = pygame.key.get_repeat()
        pygame.key.set_repeat(100, 50)
        
    def handle_focus_lost(self, **kwargs):
        pygame.key.set_repeat(*self._orig_key_repeat)
        self.validate()
        # text could still contain the cursor
        self.text = self._text
                
         
    def handle_events(self, events):
        # unused events
        not_used = []
        text = self._text
        if self.cursor_pos > len(text):
            self.cursor_pos = len(text)
        
        if self.focused:
            for event in events:
                if event.type == loc.KEYDOWN:
                    if event.key == loc.K_BACKSPACE:
                        if self.cursor_pos == 0:
                            continue
                        text = text[:self.cursor_pos-1] + text[self.cursor_pos:]
                        self.cursor_pos -= 1
                        if self.cursor_pos < 0:
                            self.cursor_pos = 0
                    elif event.key == loc.K_DELETE:
                        text = text[:self.cursor_pos] + text[self.cursor_pos+1:]
                    elif event.key == loc.K_LEFT:
                        self.cursor_pos -= 1
                        if self.cursor_pos < 0:
                            self.cursor_pos = 0
                    elif event.key == loc.K_RIGHT:
                        self.cursor_pos += 1
                        if self.cursor_pos > len(text):
                            self.cursor_pos = len(text)
                    elif event.key == loc.K_HOME:
                        self.cursor_pos = 0
                    elif event.key == loc.K_END:
                        self.cursor_pos = len(text)
                    elif event.key == loc.K_INSERT:
                        self.insert = not self.insert
                    elif event.key in (loc.K_RSHIFT, loc.K_LSHIFT):
                        pass
                    elif (event.key in (loc.K_RETURN, loc.K_KP_ENTER) 
                          and not self.multiline):
                        self._send_event("pressed_return")
                    else:
                        i = 0 if self.insert else 1
                        text = (text[:self.cursor_pos] +  event.unicode
                                + text[self.cursor_pos+i:])
                        self.cursor_pos += 1
                else:
                    not_used.append(event)
            if self.focused:
                self.text = (text[:self.cursor_pos] + self.cursor
                             + text[self.cursor_pos:])
                self._text = text
        else:
            not_used = events
        return not_used

#    def get_styles(self):
#        styles = Text.get_styles(self)
#        return styles
    
    def update_widget(self):
        """
        If focused, add a blinking cursor to the text.
        """
        if self.focused:
            time = pygame.time.get_ticks()
            if time - self._time > self.interval * 2:
                self._time = time
            elif time -  self._time > self.interval:
                self.cursor = " "
            else:
                self.cursor = "|" if self.insert else "_"
        #else:
        #    self.text = self._text
        
        
    def validate(self):
        """
        Validate input data and maybe correct it.
        """
        pass
    
    def get_text(self):
        return self._text
    
    def set_text(self, text):
        self.text = self._text = text
      
        
class NumberInput(TextInput):
    multiline = False
    def __init__(self, *args, **kwargs):
        """
        Extra arguments:
        :type convert_func: function
        :param convert_func: 
            Used to convert the string input to a number
        :type min_val: number or None
        :param min_val: Minimum value, used for validation.
        :type max_val: number or None
        :param max_val: Maximum value, used for validation.

        Extra events:
        
        **number_changed** (old, new)
          Number changed from ``old`` to ``new``

        """
        self.convert_func = kwargs.get("convert_func", int)
        self.min_val = kwargs.get("min_val", None)
        if self.min_val: 
            self.min_val = self.convert_func(self.min_val)
        self.max_val = kwargs.get("max_val", None)    
        if self.max_val: 
            self.max_val = self.convert_func(self.max_val)
        TextInput.__init__(self, *args, **kwargs)
        self.number = 0
        self.validate()
        
    def validate(self):
        try:
            old = self.number
            self.number = self.convert_func(self.get_text(validate=False))
            if self.min_val is not None:
                self.number = max(self.min_val, self.number)
            if self.max_val is not None:
                self.number = min(self.max_val, self.number)
            if self.number != old:
                self._send_event("number_changed", new=self.number, old=old)
        except ValueError:
            pass
        self.set_number(self.number, validate=False)
            
    def get_text(self, validate=True):
        if validate:
            self.validate()
        return TextInput.get_text(self)
    
    def get_number(self):
        self.validate()
        return self.number
    
    def set_number(self, number, validate=False):
        self.set_text(str(number), validate)
        
    def set_text(self, text, validate=True):
        TextInput.set_text(self, text)
        if validate:
            self.validate()





class AdvancedNumberInput(NumberInput):
    focusable = False
    def __init__(self, *args, **kwargs):
        """
        Extra arguments:
        :type step_size: number
        :param step_size: Amount to add per pixel on mouse_dragged.
        """
        self.step_size = kwargs.get("step_size", 1)    
        NumberInput.__init__(self, *args, **kwargs)
        self.init_number = None

    def handle_mouse_dragged(self, abs, rel, buttons):
        if self.init_number is None:
            self.init_number = self.number
        
        self.set_number(self.convert_func(self.init_number +
                                          rel[0]*self.step_size))
        self.text = self._text
        
    def handle_mouse_stopped_dragging(self, pos):
        self.init_number = None
        
        
    
     

    
class ScrolledLayer(Layer):
    style_class_name = "layer"
    def __init__(self, parent, size=None, pos=(0, 0), scroll_size=None, **kwargs):
        """
        Scrollable Layer.
        Not yet for productive use!
        
        Use the scroll_size property to set the scrolled area (ie. what you can
        reach by scrolling)
        """
        Layer.__init__(self, parent, size, pos, **kwargs)
        self.init = False
        self._offset_x, self._offset_y = 0, 0
        self._cached_offsets = 0, 0
        # don't scroll these widgets
        self.fixed_widgets = []
 
        self._scroll_size = (10000000,10000000)
        self.scroll_size = scroll_size or self.size
        self.init = True

        
    def scroll_x(self, offset):
        """
        :type offset: int
        :param offset:
            amount to scroll to the right
            
        Scroll the layer horizontally.
        """
        self.offset_x += offset
        
    def scroll_rel_x(self, amount):
        """
        :type amount: float or int
        :param amount:
            relative amount to scroll to the right
            
        Scroll the layer horizontally.
        """
        scroll_width = self.scroll_size[0]
        visible_width = self.get_size("scrollarea")[0]
        scroll_length = scroll_width - visible_width
        self.offset_x += - amount * scroll_length / 100.

    def scroll_to_x(self, new_x):
        """
        :type new_x: int
        :param new_x:
            Absolute horizontal position to scroll to
        """
        self.offset_x = new_x
        
    def scroll_to_rel_x(self, p):
        """
        :type p: int
        :param p:
            Relative horizontal position, between 0 and 100
        """
        scroll_width = self.scroll_size[0]
        visible_width = self.get_size("scrollarea")[0]
        scroll_length = scroll_width - visible_width
        self.offset_x = - p * scroll_length / 100.
        
        
    def scroll_y(self, offset):
        """
        :type offset: int
        :param offset:
            amount to scroll down
            
        Scroll the layer vertically.
        """
        self.offset_y += offset

    def scroll_rel_y(self, amount):
        """
        :type amount: float or int
        :param amount:
            relative amount to scroll down
            
        Scroll the layer vertically.
        """
        scroll_height = self.scroll_size[1]
        visible_height = self.get_size("scrollarea")[1]
        scroll_length = scroll_height - visible_height
        self.offset_y += - amount * scroll_length / 100.
        
    def scroll_to_y(self, new_x):
        """
        :type new_y: int
        :param new_y:
            Absolute vertical position to scroll to
        """
        self.offset_y = new_x
        
    def scroll_to_rel_y(self, p):
        """
        :type p: int
        :param p:
            Relative vertical position, between 0 and 100
        """
        scroll_height = self.scroll_size[1]
        visible_height = self.get_size("scrollarea")[1]
        scroll_length = scroll_height - visible_height
        self.offset_y = - p * scroll_length / 100.
        
    def init_pos(self):
        """
        Reset all scrolling offsets.
        """
        self.offset_x, self.offset_y = 0, 0
        self._send_event("init_pos")

#    def get_rect(self, widget):
#        rect = Layer.get_rect(self, widget)
#        if widget not in self.fixed_widgets:
#            rect.move_ip(self.offset_x, self.offset_y)
#        return rect
    
    def get_pos(self, widget):
        rect = self.get_rect(widget)
        if widget not in self.fixed_widgets:
            rect.move_ip(self.offset_x, self.offset_y)
        return rect.topleft
    
    def get_size(self, widget):
        rect = self.get_rect(widget)
        if widget not in self.fixed_widgets:
            rect.move_ip(self.offset_x, self.offset_y)
        return rect.size

#    def get_styles(self):
#        styles = Layer.get_styles(self)
#        styles["offset"] = (self.offset_x, self.offset_y)
#        #! ends in circular event calls
#        #if self._cached_offsets != (self.offset_x, self.offset_y):
#        #    self._send_event("new_offsets")
#        return styles
    
    def get_scroll_size(self):
        """
        size of the scrollable area. Should be larger than self.size
        """
        return self._scroll_size
    
    def set_scroll_size(self, new):
        self._scroll_size = max(new[0], self.get_size("scrollarea")[0]), max(new[1], self.get_size("scrollarea")[1])
        self._send_event("new_scroll_size")
        
    scroll_size = property(fget=get_scroll_size, fset=set_scroll_size)
    
    def _get_off_x(self):
        return self._offset_x
    
    def _set_off_x(self, new_x):
        self._offset_x = max(-self.scroll_size[0]+self.get_size("scrollarea")[0], min(0, new_x))
#        print "set off x", new_x, self.get_size("scrollarea")[0], self.get_size("scrollarea")[0], self.offset_x
        self.dirty = True
        self._send_event("new_x_offset")
        
    def _get_off_y(self):
        return self._offset_y
    
    def _set_off_y(self, new_y):
        self._offset_y = max(-self.scroll_size[1]+self.get_size("scrollarea")[1], min(0, new_y))
        self.dirty = True
        self._send_event("new_y_offset")
    
    offset_x = property(_get_off_x, _set_off_x)    
    offset_y = property(_get_off_y, _set_off_y)
    
                                                    
#    def handle_mouse_down(self, button, **kwargs):
#        if button == 4:
#            self.scroll_x(10)
#        elif button == 5:
#            self.scroll_x(-10)
#        else:
#            print button

class Slider(Widget):
    pass

class Scrollbar(Layer):
    def __init__(self, parent, size, pos=(0, 0), orient="y", auto_toggle=True,
                **kwargs):
        Layer.__init__(self, parent, size, pos, **kwargs)
        parent.fixed_widgets.append(self)
        # hide when not needed
        self.auto_toggle = auto_toggle
        # orient: either x/0 or y/1 for x or y values
        if orient in ("x", 0):
            self.orient = 0
            
            self.slider = Slider(self, (0, 0), ("0%", 0), 
                             validate_pos=True, 
                             style=Style(fallback=self.style))
            self.parent.listen("new_x_offset", self.update_slider_pos)
        elif orient in ("y", 1):
            self.orient = 1
            self.slider = Slider(self, (0, 0), (0, "0%"), 
                             validate_pos=True,
                             style=Style(fallback=self.style))
            self.parent.listen("new_y_offset", self.update_slider_pos)
        else:
            raise (ValueError, 
                   "Invalid orient '%s', must be one of 'x', 0, 'y' or 1" % 
                   orient)
        self.update_slider_size()
        self.slider.listen("mouse_dragged", self.handle_slider_move)
        self.parent.listen("init_pos", lambda **kw: self.init_pos())
        self.parent.listen("new_scroll_size", self.update_slider_size)
        self.parent.listen("resized", self.update_slider_size)
#        self.parent.listen("new_offsets", self.update_slider_pos)
        
    def handle_slider_move(self, rel,abs, **kwargs):
       # print "handle slider move"
#        if self.orient == 0:
#            self.slider.move((rel[0], 0))
#        else:
#            self.slider.move((0, rel[1]))
        i = self.orient
#        #rel_pos = ((self.slider.pos[i] - self.get_pos(self.slider)[i]) /
#        #           float(self.slider.size[i]))
#        #new_scroll_pos = self.parent.scroll_size[i] * -rel_pos
        slider_range = self.get_size(self.slider)[i] - self.slider.size[i]
#        rel_pos = slider_range and self.slider.local_pos[i] * 100 / slider_range
        amount = slider_range and rel[i] / slider_range * 100
        if self.orient == 0:
            #self.parent.scroll_to_x(new_scroll_pos)
            #self.parent.scroll_to_rel_x(rel_pos)
            self.parent.scroll_rel_x(amount)
        else:
            #self.parent.scroll_to_y(new_scroll_pos)
            #self.parent.scroll_to_rel_y(rel_pos)
            self.parent.scroll_rel_y(amount)
            
    def update_slider_pos(self, **kw):
        i = self.orient
        if i == 0:
            offset = self.parent.offset_x
        else:
            offset = self.parent.offset_y
#        slider_range = self.size[i]#self.get_size(self.slider)[i]# - self.slider.size[i]
        slider_range = self.get_size(self.slider)[i] - self.slider.size[i]
        start = self.slider.local_pos
#        print slider_range, offset, self.parent.scroll_size[i]
        parent_rel_pos = (self.parent.scroll_size[i]-self.parent.get_size("scrollarea")[i])
        if parent_rel_pos <= 0:
            rel_pos = 0
        else:
            rel_pos = slider_range * ((offset* -1) / parent_rel_pos) 
#        print "-->", slider_range, offset, self.parent.scroll_size[i], rel_pos
        if i == 0:
#            print rel_pos, start[1]
            self.slider.pos = rel_pos, start[1]
        else:
            self.slider.pos = start[0], rel_pos
        
            
    def init_pos(self):
        #print "init pos"
        if self.orient == 0:
            self.slider.pos = ("0%", 0)
        else:
            self.slider.pos = (0, "0%")
            
    def update_slider_size(self, **kw):
       # print "update slider size"
        #!?
        px, py = self.parent.get_size("scrollarea")#self)#!
        sx, sy = self.parent.scroll_size
#        print "update slider size", self.parent.get_size(self), self.parent.scroll_size
     #   print "parent size", px,py
     #   print "scroll size", sx,sy
#        print "update slider siiz"
        if self.orient ==  0:
            # avoid ZeroDivisionError
            i = int(px* 100./max(sx, 1))
#            print "slider i", i
            if i >= 100 and self.auto_toggle:
                self.hide()
                return
            self.slider.size = ("%i%%" % min(i, 100), "100%")
            self.show()
        else:
            i = int(py* 100./max(sy, 1))
       #     print i
            if i >= 100 and self.auto_toggle:
                self.hide()
                return
            self.slider.size = ("100%", "%i%%" % min(i, 100))
            self.show()
            
            
class Switch(Widget):
    def __init__(self, parent, size, pos, states=["on", "off"], **kwargs):
        Widget.__init__(self, parent, size, pos, **kwargs)
        self.states = states
        self.state = 0
        
    def _set_state(self, state):
        """
        :type state: int
        :param state:
            New state index.
            
        For internal use. Set a new state.
        """
        old_state = self.get_state()
        self.state = state
        new_state = self.get_state()
        self.mark_dirty()
        self._send_event("state_changed", old_state=old_state,
                        new_state=new_state)
        
    def handle_mouse_down(self, button, **kwargs):
        if button == 1:
            self._set_state((self.state + 1) % len(self.states))
        
    def get_state(self):
        return self.states[self.state]
    
    def set_state(self, state):
        if state not in self.states:
            raise ValueError, "Invalid state '%s', must be one of %s" % (
                                state, ", ".join(self.states))
        self._set_state(self.states.index(state))
        
    def get_suffix(self, name):
        if name == "image":
            return self.get_state()
    
    
#class Checkbox(Layer):
#    """
#    Switch with text
#    """
#    def __init__(self, parent, size, pos=(0, 0), value=1, text="", **kwargs):
#        Layer.__init__(self, parent, size, pos, **kwargs)
#        self.text_widget = Text(self, "auto", pos, text)
        

class TabHead(Layer):
    def __init__(self, parent, name, **kwargs):
        self.auto_size = parent.head_width == 0
        if self.auto_size:
            # just an initial value
            width = 70
        else:
            width = parent.head_width            
        height = parent.head_height
        
        Layer.__init__(self, parent, (width, height), (0, 0), **kwargs)
        self.tab_name = name
        self.text = Text(self, "auto", ("50%", 0), name, anchor="midtop",
                         pass_events=True)
        self.text.listen("resized", self.handle_text_resized)
        self.handle_text_resized()
        
    def handle_text_resized(self, **kw):
        # inner rect
        irect = self.get_rect(self.text)
        tsize = self.text.size
        if self.auto_size:
            x_border = self.rect.width - irect.width
            y_border = self.rect.height - irect.height
            self.size = (tsize[0] + x_border,
                         self.parent.head_height)
                         #tsize[1] + y_border)
        # if the text does not fit completely into the tabhead, change its 
        # position and show the beginning of the text
        if irect.width < tsize[0]:
            self.text.pos = (0, 0)
            self.text.anchor = "topleft"
        #! also do that for height (and set height to 50% as default too)?
        
    def handle_mouse_down(self, button, **kw):
        if button == 1:
            self.parent.set_active_tab(self.tab_name)
            
    def get_suffix(self, name):
        if name == "image":
            return self.is_active() and "active"
    
    def is_active(self):
        return self.tab_name == self.parent.active_tab
            
        

class TabbedLayer(Layer):
    default_head_width = 70
    default_head_height = 30
    def __init__(self, parent, size, pos=(0, 0), **kwargs):
        Layer.__init__(self, parent, size, pos, **kwargs)
        #: list of tab names
        self.order = []
        #: dict of tab name -> heads
        self.heads = {}
        #: dict of tab name -> bodies
        self.bodies = {}
        # name of active tab
        self.active_tab = ""
        #: if 0, use individual widths
        self.head_width = kwargs.get("head_width", self.default_head_width)
        #: absolute amount of pixels
        self.head_height = kwargs.get("head_height", self.default_head_height)
        
    def new_tab(self, name, index="end", body=None):
        """
        :type body: :class:`gui.base.Layer`
        :param body: 
            A Layer to use as body. This allows you to use child classes
            of Layer, eg. :class:`ScrolledLayer`
        """
        # tab names must be unique
   #     print "add tab '%s'" % name
    #    print self.order
        # if a tab with ``name`` already exists, add a numbering
        if name in self.order:
            i = 2
            while "%s %i" % (name, i) in self.order:
                i += 1
            name += " %i" % i
    #    print name
        # convert "end" index to int
        if index == "end":
            index = len(self.order)
        
#        head_size =  (self.head_width * index, "100%")
#        head_size = self.head_width and head_size or "auto"
        head = TabHead(self, name)

        #head.listen("mouse_down", self.handle_head_mouse_down)
        self.heads[name] = head
        
        if not body:
            body = Layer(self, ("100%", "100%"))
        else:
            body.size = ("100%", "100%")
            body.pos = (0, 0)
        body.tab_name = name
        self.bodies[name] = body
        
        if not self.active_tab:
            self.active_tab = name
            body.show()
        else:
            body.hide()
                
        self.order.insert(index, name)
        self.reorder()
        
        self._send_event("tab_added", name=name)   
        
        return body     
        
        
    def handle_tab_added(self, name, **kw):
        pass
            
            

            
    def set_active_tab(self, tab_name):
        if self.active_tab == tab_name:
            return
        self.bodies[self.active_tab].hide()
        old_active_tab = self.active_tab
        self.active_tab = tab_name
        self.bodies[self.active_tab].show()
        self._send_event("new_active_tab", 
                         old_active_tab=old_active_tab,
                         new_active_tab=self.active_tab)
  #      print "new active tab:", tab_name
            
            
            
    def reorder(self):
        """
        set new positions of tabs
        """
        x_count = 0
        for name in self.order:
            head = self.heads[name]
            head.pos = x_count, 0
            if self.head_width:
                head.size = self.head_width, "100%"
            x_count += head.size[0]
      #      print "reorder", x_count
      #  print
            
            
        
    def get_rect(self, widget):
        bd = self.style("borderwidth", self, 0)
        if type(bd) == int:
            bd = bd, bd
        width = self.size[0] - 2 * bd[0]
        x0, y0 = self.pos[0] + bd[0], self.pos[1] + bd[1]
        if widget in self.heads.itervalues():
            return pygame.Rect(x0, y0, width, self.head_height)
        else:
            return pygame.Rect(x0, y0 + self.head_height, width,
                               self.size[1] - 2 * bd[1] - self.head_height)


class RectWidget(Widget):
    """
    Unvisible Widget to define click areas.
    """
    pass_events = True
    def __init__(self, parent, pos, size=(30,30), anchor="topleft", **kwargs):

        Widget.__init__(self, parent, size, pos, anchor=anchor, **kwargs)

    def blit(self, screen):
        #drects = Widget.blit(self, screen)
        return #[]

class Window(Layer):
    #style_class_name = "layer"
    def __init__(self, parent, title, size=None, pos=(0, 0), **kwargs):
#        sizer_data =[
#             [("0", "0"), (bd, "100%"), "topleft", (-1, 0, "right")],
#             [("0", "0"), ("100%", bd), "topleft", (0, -1, "bottom")],
#             [("100%", "100%"), ("100%", bd), "bottomright", (0, 1, "left")],
#             [("100%", 0), (bd, "100%"), "topright"), (1, 0, "top")]
#                     ]
                     
#        for pos, size, anch, resize_param, resize_anch in sizer_data:
#            sizer = RectWidget(fg_layer, pos, size)
#            sizer.listen("mouse_dragged",
#                         lambda **kw: fg_layer.resize((-kw["abs"][0], 0), anchor="right"))
        self.sizers = []
        Layer.__init__(self, parent, size, pos, **kwargs)
        self.title = title
        bd =  self.style("borderwidth", self, 0)
        
        sizer = RectWidget(self, ("0", "0"), (bd, "100%"))
        sizer.listen("mouse_dragged",
                lambda **kw: self.resize((-kw["rel"][0], 0), anchor="right"))
        self.sizers.append(sizer)
        sizer = RectWidget(self, ("0", "0"), ("100%", bd))
        sizer.listen("mouse_dragged", 
                lambda **kw: self.resize((0, -kw["rel"][1]), anchor="bottom")
                if kw["buttons"][2] else self.move(kw["abs"]))
        self.sizers.append(sizer)
        sizer = RectWidget(self, ("100%", "100%"), ("100%", bd), 
                           anchor="bottomright")
        sizer.listen("mouse_dragged", 
                lambda **kw: self.resize((0, kw["rel"][1]), anchor="left"))
        self.sizers.append(sizer)
        sizer = RectWidget(self, ("100%", 0), (bd, "100%"), anchor="topright")
        sizer.listen("mouse_dragged", 
                lambda **kw: self.resize((kw["rel"][0], 0), anchor="top"))
        self.sizers.append(sizer)
        
    def get_rect(self, widget):
        if widget in self.sizers:
            return self.rect
        return Layer.get_rect(self, widget)